/*
 * Copyright (c) Linux community.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
#include <string.h>
#include <glib.h>
#include <gdk/gdk.h>
#include <dbus/dbus.h>
#include <dbus/dbus-glib.h>

#include "log.h"
#include "dbus-player.h"
#include "dbus-mpris.h"
#include "dbus-rhythmbox.h"
#include "dbus-banshee.h"

// Code for Media Players that strictly follow the MPRIS DBus standard.
// Ref.: http://xmms2.org/wiki/MPRIS
//
// This code is largely based on Pidgin-musictracker's code, mpris.c module.
// Copyrights and thanks to http://code.google.com/p/pidgin-musictracker/

#define DBUS_MPRIS_PLAYER         "org.freedesktop.MediaPlayer"
#define DBUS_MPRIS_ROOT_PATH      "/"
#define DBUS_MPRIS_PLAYER_PATH    "/Player"
#define DBUS_MPRIS_TRACK_SIGNAL   "TrackChange"
#define DBUS_MPRIS_STATUS_SIGNAL  "StatusChange"

// Value types for DBus messages
#define DBUS_TYPE_G_STRING_VALUE_HASHTABLE (dbus_g_type_get_map ("GHashTable", G_TYPE_STRING, G_TYPE_VALUE))
#define DBUS_TYPE_MPRIS_STATUS (dbus_g_type_get_struct ("GValueArray", G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_INVALID))

void mpris_status_signal_int_cb(DBusGProxy *player_proxy, gpointer data, MediaPlayerRec *player);
void mpris_status_signal_struct_cb(DBusGProxy *player_proxy, gpointer data, MediaPlayerRec *player);

void mpris_module_init() {
    LOG_DEBUG("Init dbus_mpris.c.\n");
}

void mpris_module_exit() {
    LOG_DEBUG("Clean up dbus_mpris.c.\n");
}

void mpris_player_track_changed(MediaPlayerRec *player) {
    // Called when the song/audio track changes

    // Re-read all player data
    mpris_get_info(player);

    // Set track changed signal
    TrackInfo *tr = &player->track;
    tr->status = PLAYER_STATUS_TRACK_CHANGE;

    // Print debug info
    // dbus_player_debug_print(player);

    // Process data
    dbus_player_process_data(player);

}

void mpris_player_status_changed(MediaPlayerRec *player) {
    // Called when player's status changes

    // Get player information
    mpris_get_info(player);

    // Print debug info
    // dbus_player_debug_print(player);

    // Process data
    dbus_player_process_data(player);
}

gboolean mpris_service_is_running(gpointer player_rec) {
    // Check if media player app is is running/is active
    MediaPlayerRec *player = (MediaPlayerRec*)player_rec;
    return mpris_service_is_running_by_name(player->service_name);
}

gboolean mpris_service_is_running_by_name(const char *service_name) {
    // Check if given service (media player application) is running/is active
    // ref.: http://www.koders.com/c/fid8FD8DF8B9ED4F45A30476317D5A3261237576BAA.aspx?s=%22Christian+Mueller%22

    DBusGProxy *proxy = NULL;
    GError *error = NULL;

    DBusGConnection *dbus_conn = dbus_player_connect_to_dbus();

    proxy = dbus_g_proxy_new_for_name(dbus_conn,
                                      DBUS_SERVICE_DBUS,   // "org.freedesktop.DBus"
                                      DBUS_PATH_DBUS,      // "org/freedesktop/DBus"
                                      DBUS_INTERFACE_DBUS);// "org.freedesktop.DBus"
    if (!proxy) {
        return FALSE;
    }

    gboolean running = FALSE;

    if (dbus_g_proxy_call_with_timeout(proxy, "NameHasOwner", DBUS_MPRIS_TIMEOUT, &error,
                                       G_TYPE_STRING, service_name,
                                       G_TYPE_INVALID,
                                       G_TYPE_BOOLEAN, &running,
                                       G_TYPE_INVALID)) {
    }

    if (error) {
        LOG_ERROR("mpris_service_is_running_by_name: NameHasOwner call for %s failed. %s.\n", service_name, error->message);
        g_error_free(error);
    }

    // Unref the proxy
    g_object_unref(proxy);

    return running;
}

void mpris_read_track_info(GHashTable *table, MediaPlayerRec *player) {
    GValue *value;
    TrackInfo *tr = &player->track;

    // Fetch values from the hash table
    // Artist
    value = (GValue *) g_hash_table_lookup(table, "artist");
    if (G_VALUE_HOLDS_STRING(value)) {
        g_strlcpy(tr->artist, g_value_get_string(value), MPRIS_STRLEN-1);
    }

    // Album
    value = (GValue *) g_hash_table_lookup(table, "album");
    if (G_VALUE_HOLDS_STRING(value)) {
        g_strlcpy(tr->album, g_value_get_string(value), MPRIS_STRLEN-1);
    }

    // Title
    value = (GValue *) g_hash_table_lookup(table, "title");
    if (G_VALUE_HOLDS_STRING(value)) {
        g_strlcpy(tr->track, g_value_get_string(value), MPRIS_STRLEN-1);
    }

    tr->total_secs = 0;

    // Time values
    value = (GValue *)g_hash_table_lookup(table, "time");

    if (G_VALUE_HOLDS_UINT(value))
        tr->total_secs =  g_value_get_uint(value);
    else if (G_VALUE_HOLDS_UINT64(value))
        tr->total_secs =  g_value_get_uint64(value);
    else if (G_VALUE_HOLDS_INT64(value))
        tr->total_secs =  g_value_get_int64(value);

    // Note:
    // The hash map contains probably other keys, like: "location", "tracknumber", "genre", etc.
}

void mpris_get_info(gpointer player_rec) {
    MediaPlayerRec *player = (MediaPlayerRec*) player_rec;
    TrackInfo *tr = &player->track;

    // Is this player running/is active?
    player->active = mpris_service_is_running_by_name(player->service_name);
    if (!player->active) {
        tr->status = PLAYER_STATUS_CLOSED;
        return;
    }

    GError *error = NULL;
    GHashTable *table = NULL;

    if(dbus_g_proxy_call_with_timeout(player->proxy,
                                      "GetMetadata",
                                      DBUS_MPRIS_TIMEOUT, &error,
                                      G_TYPE_INVALID, DBUS_TYPE_G_STRING_VALUE_HASHTABLE, &table,
                                      G_TYPE_INVALID)) {

        // Read values from the hash table
        mpris_read_track_info(table, player);

        // Free it
        g_hash_table_destroy(table);
    }

    if (error) {
        // The player has shutdown
        tr->status = PLAYER_STATUS_CLOSED;
        player->active = FALSE;

        LOG_ERROR("mpris_get_info: GetMetadata call for %s failed. %s\n", player->service_name, error->message);
        g_error_free(error);
        return;
    }

    GValueArray *s = NULL;
    error = NULL;
    if (dbus_g_proxy_call_with_timeout(player->proxy, "GetStatus", DBUS_MPRIS_TIMEOUT, &error,
                                       G_TYPE_INVALID,
                                       DBUS_TYPE_MPRIS_STATUS, &s,
                                       G_TYPE_INVALID)) {

        // Read player's status
        gint status = PLAYER_STATUS_CLOSED;

        if (s) {
            GValue *v = g_value_array_get_nth(s, 0);
            status = g_value_get_int(v);
        }

        g_value_array_free(s);

        // Ref: http://xmms2.org/wiki/MPRIS#GetStatus
        // 0 = Playing, 1 = Paused, 2 = Stopped.

        // Convert player status to this program's value system
        switch (status) {
        case 0:
            tr->status = PLAYER_STATUS_PLAYING;
            break;

        case 1:
            tr->status = PLAYER_STATUS_PAUSED;
            break;

        case 2:
            tr->status = PLAYER_STATUS_STOPPED;
            break;

        default:
            tr->status = PLAYER_STATUS_CLOSED;
        }
    }

    if (error) {
        tr->status = PLAYER_STATUS_CLOSED;
        LOG_ERROR("mpris_get_info: GetStatus call for %s failed. %s\n", player->service_name, error->message);
        g_error_free(error);
    }


    // Current play time
    tr->current_secs = 0;
    gint position = 0;
    error = NULL;

    if(dbus_g_proxy_call_with_timeout(player->proxy, "PositionGet", DBUS_MPRIS_TIMEOUT, &error,
                                      G_TYPE_INVALID, G_TYPE_INT, &position,
                                      G_TYPE_INVALID)) {
        tr->current_secs = position/1000;
    }

    if (error) {
        g_error_free(error);
    }

}

void debug_hash_table(MediaPlayerRec *player, GHashTable *table) {

    GHashTableIter iter;
    g_hash_table_iter_init(&iter, table);

    gpointer key, value;

    g_hash_table_iter_init(&iter, table);
    while (g_hash_table_iter_next(&iter, &key, &value)) {
        // The keys defined by Totem are:
        // "title", "location", "time", "genre", "album", "tracknumber", "year" and "artist"
        //
        // The keys defined by Amarok are:
        // "comment", "genre", "location", "rating", "arturl", "album", "audio-bitrate", "tracknumber", "year", "artist", "audio-samplerate" and "title"

        glong long_val = G_MINLONG;
        if (G_VALUE_HOLDS_STRING(value)) {
            LOG_PLAYER("%s %s: %s=%s\n", player->service_name, player->app_name, key, g_value_get_string(value));
        } else if G_VALUE_HOLDS_UINT(value) {
            long_val = (glong)g_value_get_uint(value);
        } else if G_VALUE_HOLDS_INT(value) {
            long_val = (glong)g_value_get_int(value);
        } else if G_VALUE_HOLDS_LONG(value) {
            long_val = (glong)g_value_get_long(value);
        } else if G_VALUE_HOLDS_ULONG(value) {
            long_val = (glong)g_value_get_ulong(value);
        }

        if (long_val != G_MINLONG) {
            LOG_PLAYER("%s %s: %s=%ld\n", player->service_name, player->app_name, key, long_val);
        }
    }
    LOG_PLAYER("------------------------\n");
}


void mpris_track_signal_cb(DBusGProxy *player_proxy, gpointer data, MediaPlayerRec *player) {
    GHashTable *hash_table = (GHashTable*)data;
    if (hash_table) {
        /*suppress unused variable msg*/
    }

    LOG_PLAYER("mpris_track_signal_cb()\n");

#ifdef ACTIVE_DEBUGGING
    debug_hash_table(player, hash_table);
#endif

    // Re-read all player data and inform GUI
    mpris_player_track_changed(player);
}

void mpris_status_signal_int_cb(DBusGProxy *player_proxy, gpointer data, MediaPlayerRec *player) {
    gint status = (gint)(glong)data;
    if (status) {
        /*suppress unused variable msg*/
    }

    LOG_PLAYER("mpris_status_signal_int_cb: Status: %d\n", status);

    // Re-read all player data and inform the GUI
    mpris_player_status_changed(player);
}

void mpris_status_signal_struct_cb(DBusGProxy *player_proxy, gpointer data, MediaPlayerRec *player) {
    GValueArray *sigstruct = (GValueArray*)data;

    gint status = -1;
    if (sigstruct) {
        GValue *v = g_value_array_get_nth(sigstruct, 0);
        status = g_value_get_int(v);
    }

    // Suppress "not used" message
    if (status) {
        ;
    }

    LOG_PLAYER("mpris_status_signal_struct_cb(): status=%d\n", status);

    // Re-read all player data and inform GUI
    mpris_player_status_changed(player);
}

void mpris_set_signals(gpointer player_rec, gboolean connect) {
    // Connect/disconnect signals

    MediaPlayerRec *player = (MediaPlayerRec*)player_rec;

    DBusGConnection *dbus_conn = dbus_player_connect_to_dbus();

    if (!player->proxy) {
        player->proxy = dbus_g_proxy_new_for_name(dbus_conn, player->service_name, DBUS_MPRIS_PLAYER_PATH, DBUS_MPRIS_PLAYER);

        // Register arguments.
        dbus_g_proxy_add_signal(player->proxy, DBUS_MPRIS_TRACK_SIGNAL, DBUS_TYPE_G_STRING_VALUE_HASHTABLE, G_TYPE_INVALID);
        dbus_g_proxy_add_signal(player->proxy, DBUS_MPRIS_STATUS_SIGNAL, DBUS_TYPE_MPRIS_STATUS, G_TYPE_INVALID);
    }

    if (!player->proxy) {
        LOG_ERROR("mpris_set_signals: Cannot create proxy for %s.\n", player->service_name);
        return;
    }

    if (connect) {
        // Connect signals
        dbus_g_proxy_connect_signal(player->proxy, DBUS_MPRIS_TRACK_SIGNAL, G_CALLBACK(mpris_track_signal_cb), (gpointer)player, NULL);
        dbus_g_proxy_connect_signal(player->proxy, DBUS_MPRIS_STATUS_SIGNAL, G_CALLBACK(mpris_status_signal_struct_cb), (gpointer)player, NULL);

    } else {
        // Disconnect signals
        dbus_g_proxy_disconnect_signal(player->proxy, DBUS_MPRIS_TRACK_SIGNAL, G_CALLBACK(mpris_track_signal_cb), (gpointer)player);
        dbus_g_proxy_disconnect_signal(player->proxy, DBUS_MPRIS_STATUS_SIGNAL, G_CALLBACK(mpris_status_signal_struct_cb), (gpointer)player);
    }
}

void mpris_get_app_name(MediaPlayerRec *player) {

    DBusGConnection *dbus_conn = dbus_player_connect_to_dbus();

    // Get identity/name of this player (the name + version)
    // Ref:http://xmms2.org/wiki/MPRIS#Identity

    LOG_PLAYER("mpris_get_app_name: %s\n", player->app_name);

    DBusGProxy *proxy = dbus_g_proxy_new_for_name(dbus_conn,
                        player->service_name,
                        DBUS_MPRIS_ROOT_PATH,
                        DBUS_MPRIS_PLAYER);

    gchar *player_name = NULL;
    GError *error = NULL;
    if (dbus_g_proxy_call_with_timeout(proxy, "Identity", DBUS_MPRIS_TIMEOUT, &error, G_TYPE_INVALID,  G_TYPE_STRING, &player_name, G_TYPE_INVALID)) {
        player->app_name = g_strdup(player_name);
        // g_free(player_name);
    }

    if (error) {
        LOG_ERROR("mpris_get_app_name: Cannot get identity/name for %s. %s\n", player->service_name, error->message);
        g_error_free(error);
    }
}

void mpris_start_app(gpointer player_rec) {
    MediaPlayerRec *player = (MediaPlayerRec*)player_rec;

    LOG_PLAYER("mpris_start_app: %s\n", player->app_name);

    DBusGConnection *dbus_conn = dbus_player_connect_to_dbus();

    if (!player->proxy) {
        player->proxy = dbus_g_proxy_new_for_name(dbus_conn, player->service_name,
                        DBUS_MPRIS_PLAYER_PATH,
                        DBUS_MPRIS_PLAYER);
    }

    // Start it
    dbus_g_proxy_call_no_reply(player->proxy, "Play", G_TYPE_INVALID);
}

MediaPlayerRec *mpris_player_new(const gchar *service_name) {
    // New MediaPlayerRec record
    MediaPlayerRec *player = g_malloc0(sizeof(MediaPlayerRec));
    player->service_name = g_strdup(service_name);
    player->app_name = NULL;
    return player;
}

