/*
 * Copyright (c) Linux community.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
#include <string.h>
#include <glib.h>
#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <dbus/dbus.h>
#include <dbus/dbus-glib.h>

#include "log.h"
#include "utility.h"
#include "support.h"
#include "dbus-player.h"

#include "audio-sources.h"

// Module for MPRIS compliant (standards following) players.
#include "dbus-mpris.h"

// Handcrafted DBus interfaces for various Media Players, Skype, etc.
#include "dbus-rhythmbox.h"
#include "dbus-banshee.h"
#include "dbus-skype.h"

// Glib/DBus connection
static DBusGConnection *g_dbus_conn = NULL;

// List of players
static GHashTable *g_player_list = NULL;

// Hopefully the core system sets this function
static DBusDataCallbackFunc g_data_callback_func;

static void dbus_player_disconnect_signals();
static void dbus_player_clear_list();

void dbus_player_detect_software();

static void add_rhythmbox_player();
static void add_banshee_player();

void dbus_player_init() {
    LOG_DEBUG("Init dbus-player.c.\n");

    g_player_list = NULL;
    g_dbus_conn = NULL;

    skype_module_init();

    mpris_module_init();
}

void dbus_player_exit() {
    LOG_DEBUG("Clean up dbus-player.c.\n");

    // Disconnect DBus signals for all Media Players
    dbus_player_disconnect_signals();

    // Disconnect g_dbus_conn from the DBus
    dbus_player_disconnect_from_dbus();

    // Clear the player list
    dbus_player_clear_list();

    mpris_module_exit();

    skype_module_exit();
}

void dbus_player_set_callback_func(DBusDataCallbackFunc data_callback_func) {
    // Set data callback function so we can send messages to rec-manager.c
    g_data_callback_func = data_callback_func;
}

void dbus_player_process_data(gpointer player) {
    // Send message to the rec-manager.c?

    // Has callback function?
    if (g_data_callback_func) {
        // Send
        g_data_callback_func(player);
    }
}

void dbus_player_player_changed(gchar *service_name) {
    // Re-connect DBus signals/methods for the given service_name (Media Player or Skype, etc)

    // Disconnect all signals/object methods
    dbus_player_disconnect_signals();

    // Get MediaPlayerRec for this service_name
    MediaPlayerRec *player = dbus_player_lookup_service_name(service_name);

    if (player && player->func_set_signals) {

        LOG_PLAYER("Connect DBus signals for %s (%s).\n", player->app_name, player->service_name);

        // Start application (Media Player, Skype, etc)
        if (player->func_start_app) {
            player->func_start_app(player);
        }

        // Connect signals so we receive track-changed/start/stop messages from this app (over DBus)
        player->func_set_signals(player, TRUE); // TRUE=connect/register, FALSE=disconnect/unregister

    }
}

static void dbus_player_disconnect_signals() {
    // Disconnect all signal-functions from the DBus
    GHashTableIter iter;
    gpointer key, value;

    g_hash_table_iter_init(&iter, g_player_list);
    while (g_hash_table_iter_next(&iter, &key, &value)) {
        MediaPlayerRec *player = (MediaPlayerRec*)value;
        if (player && player->func_set_signals) {
            // Disconnect signals
            player->func_set_signals(player, FALSE); // TRUE=connect/register, FALSE=disconnect/unregister
        }
    }
}

GHashTable *dbus_player_get_list_ref() {
    if (!g_player_list)
        g_player_list = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, dbus_player_delete_item);

    return g_player_list;
}

static gboolean dbus_palyer_remove_node(gpointer key, gpointer value, gpointer user_data) {
    // Return TRUE so this (key, value) pair gets removed and deleted.
    return TRUE;
}

static void dbus_player_clear_list() {
    // Delete the entire g_player_list
    if (g_player_list) {
        g_hash_table_foreach_remove(g_player_list, dbus_palyer_remove_node, NULL);
        g_hash_table_unref(g_player_list);
    }
    g_player_list = NULL;
}

void dbus_player_delete_item(gpointer data) {
    MediaPlayerRec *player = (MediaPlayerRec*)data;
    if (!player) return;

    LOG_PLAYER("dbus_player_delete_item: %s (%s).\n", player->app_name, player->service_name);

    if (player->func_set_signals) {
        // Disconnect signals
        player->func_set_signals(player, FALSE); // TRUE=connect, FALSE=disconnect
    }

    if (G_IS_OBJECT(player->proxy)) {
        g_object_unref(player->proxy);
    }
    player->proxy = NULL;

    g_free(player->service_name);
    g_free(player->app_name);
    g_free(player->icon_name);
    g_free(player);
}

DBusGConnection *dbus_player_connect_to_dbus() {
    // Connect to glib/DBus
    GError *error = NULL;

    if (!g_dbus_conn) {
        g_dbus_conn = dbus_g_bus_get(DBUS_BUS_SESSION, &error);

        if (!g_dbus_conn) {
            LOG_ERROR("dbus_player_connect_to_dbus: Cannot connect to DBus: %s\n", error ? error->message : "");

            if (error)
                g_error_free(error);

            return NULL;
        }
    }
    return g_dbus_conn;
}

void dbus_player_disconnect_from_dbus() {
    // Disconnect from glib/DBus
    if (g_dbus_conn)
        dbus_g_connection_unref(g_dbus_conn);

    g_dbus_conn = NULL;
}

MediaPlayerRec *dbus_player_lookup_app_name(gchar *app_name) {
    // Lookup player by its application name (p->app_name).
    // Typical app_names are "Amarok 2.3.2", "RhythmBox 2.3" and "Skype 2.1.0".
    GHashTableIter iter;
    gpointer key, value;

    if (str_length(app_name, 1024) < 1) return NULL;

    g_hash_table_iter_init(&iter, g_player_list);
    while (g_hash_table_iter_next(&iter, &key, &value)) {
        MediaPlayerRec *p = (MediaPlayerRec*)value;
        if (!g_strcmp0(p->app_name, app_name)) {
            return p;
        }
    }
    return NULL;
}

MediaPlayerRec *dbus_player_lookup_service_name(gchar *service_name) {
    // Lookup player by its service name (p->service_name).
    // Typical service_names are "org.gnome.Rhythmbox" and "com.Skype.API"
    GHashTableIter iter;
    gpointer key, value;

    if (str_length(service_name, 1024) < 1) return NULL;

    g_hash_table_iter_init(&iter, g_player_list);
    while (g_hash_table_iter_next(&iter, &key, &value)) {
        MediaPlayerRec *p = (MediaPlayerRec*)value;
        if (!g_strcmp0(p->service_name, service_name)) {
            return p;
        }
    }
    return NULL;
}

void dbus_player_debug_print(MediaPlayerRec *p) {
    LOG_PLAYER("------------------------------\n");

    LOG_PLAYER("Player app name:%s\n", p->app_name);
    LOG_PLAYER("Service name:%s\n", p->service_name);

    LOG_PLAYER("Player is active: %s\n", p->active ? "TRUE" : "FALSE");

    LOG_PLAYER("Flags:%d\n", p->flags);

    TrackInfo *tr = &p->track;

    switch (tr->status) {
    case PLAYER_STATUS_CLOSED:
        LOG_PLAYER("Status:%d  PLAYER_STATUS_CLOSED\n", tr->status);
        break;

    case PLAYER_STATUS_STOPPED:
        LOG_PLAYER("Status:%d  PLAYER_STATUS_STOPPED\n", tr->status);
        break;

    case PLAYER_STATUS_PAUSED:
        LOG_PLAYER("Status:%d  PLAYER_STATUS_PAUSED\n", tr->status);
        break;

    case PLAYER_STATUS_PLAYING:
        LOG_PLAYER("Status:%d  PLAYER_STATUS_PLAYING\n", tr->status);
        break;

    case PLAYER_STATUS_TRACK_CHANGE:
        LOG_PLAYER("Status:%d  PLAYER_STATUS_TRACK_CHANGE\n", tr->status);
        break;

    case PLAYER_STATUS_NOTIFY_MSG:
        LOG_PLAYER("Status:%d  PLAYER_STATUS_NOTIFY_MSG\n", tr->status);
        break;

    default:
        LOG_PLAYER("Unknown status:%d\n", tr->status);
    }

    if (tr->status != PLAYER_STATUS_NOTIFY_MSG) {
        LOG_PLAYER("Track:%s\n", tr->track);
        LOG_PLAYER("Artist:%s\n", tr->artist);
        LOG_PLAYER("Album:%s\n", tr->album);
        LOG_PLAYER("Current position in seconds:%d\n", tr->current_secs);
        LOG_PLAYER("Total length in seconds:%d\n", tr->total_secs);
        LOG_PLAYER("Flags:%d\n", tr->flags);
    } else {
        LOG_PLAYER("Message:%s\n", tr->track);
    }

    LOG_PLAYER("------------------------------\n");
}

GHashTable *dbus_player_get_player_list() {
    // Clear the old list
    dbus_player_clear_list();

    // Connect to glib/DBus
    if (!dbus_player_connect_to_dbus()) {
        LOG_ERROR("dbus_player_get_players: Error, cannot connect to DBus.\n");
        return NULL;
    }

    // Populate the list.
    // Detect players that follow the MPRIS* standarden, such as Amarok.
    // Add also non-standard media players like Banshee and RhythmBox.
    // Add also Skype if it's installed.
    dbus_player_detect_software();

    // Add Rhythmbox manually (if installed)
    add_rhythmbox_player(TRUE);

    // Add Banshee manually (if installed)
    add_banshee_player(TRUE);

    // Add Totem (if installed)
    add_totem_player(TRUE);

    // Add Skype (if installed)
    add_skype(TRUE);

    return g_player_list;
}

void add_skype(gboolean find) {
    // Add Skype

    gchar *service_name = "com.Skype.API";

    // Already in the list?
    if (dbus_player_lookup_service_name(service_name)) return;

    if (find) {
        // Find /usr/bin/skype
        gchar *path = find_command_path("skype");
        if (!path) {
            return;
        }
        g_free(path);
    }

    // New MediaPlayer record (yep, we store Skype data in a MediaPlayer record)
    MediaPlayerRec *player = mpris_player_new(service_name);

    player->type = COMM_PROGRAM;

    // Set application name
    skype_get_app_name(player);

    // Function to register/unregister notification methods
    player->func_set_signals = skype_setup;

    // Function to get track-info (strictly, Skype does not have "track" data, though it has target filename)
    player->func_get_info = skype_get_info;

    // Function to start/run Skype
    player->func_start_app = skype_start_app;

    // Set icon
    player->icon_name = g_strdup("skype.png");

    if (!dbus_player_lookup_app_name(player->app_name)) {
        // Add to list
        GHashTable *player_list = dbus_player_get_list_ref();
        g_hash_table_insert(player_list, g_strdup(player->service_name), player);
    } else {
        // Duplicate or bad record. Free it.
        dbus_player_delete_item(player);
    }
}

void add_rhythmbox_player(gboolean find) {
    // Add RhythmBox

    gchar *service_name = "org.gnome.Rhythmbox";

    // Already in the list?
    if (dbus_player_lookup_service_name(service_name)) return;

    if (find) {
        // Find /usr/bin/rhythmbox
        gchar *path = find_command_path("rhythmbox");
        if (!path) {
            return;
        }
        g_free(path);
    }

    // New MediaPlayer record
    MediaPlayerRec *player = mpris_player_new(service_name);

    // Set real application name (player->app_name)
    // Rhythmbox does not understand the "Identity" message

    // Set real application name
    rhythmbox_get_app_name(player);

    // Function to connect/disconnect event signals for this player
    player->func_set_signals = rhythmbox_set_signals;

    // Function to get track-info (album, track/song name/title, genre, etc.)
    player->func_get_info = rhythmbox_get_info;

    // Function to start/run the media player
    player->func_start_app = rhythmbox_start_app;

    // Function to check if this player is up & running
    player->active = mpris_service_is_running_by_name(service_name);

    // Set icon
    player->icon_name = g_strdup("rhythmbox.png");

    if (!dbus_player_lookup_app_name(player->app_name)) {
        // Add to list
        GHashTable *player_list = dbus_player_get_list_ref();
        g_hash_table_insert(player_list, g_strdup(player->service_name), player);
    } else {
        // Duplicate or bad record. Free it.
        dbus_player_delete_item(player);
    }
}

static void add_banshee_player(gboolean find) {
    // Add Banshee

    gchar *service_name = "org.bansheeproject.Banshee";

    // Already in the list?
    if (dbus_player_lookup_service_name(service_name)) return;

    if (find) {
        // Find /usr/bin/banshee
        gchar *path = find_command_path("banshee");
        if (!path) {
            return;
        }
        g_free(path);
    }

    // New MediaPlayer record
    MediaPlayerRec *player = mpris_player_new(service_name);

    // Set real application name (player->app_name)
    // Banshee does not understand the "Identity" message
    banshee_get_app_name(player);

    // Function that can connect/disconnect event signals
    player->func_set_signals = banshee_set_signals;

    // Function to get track-info (album, track/song name/title, genre, etc.)
    player->func_get_info = banshee_get_info;

    // Function to start/run the media player
    player->func_start_app = banshee_start_app;

    // Function to check if this player is up & running
    player->active = mpris_service_is_running_by_name(service_name);

    // Set icon
    player->icon_name = g_strdup("banshee.png");

    if (!dbus_player_lookup_app_name(player->app_name)) {
        // Add to list
        GHashTable *player_list = dbus_player_get_list_ref();
        g_hash_table_insert(player_list, g_strdup(player->service_name), player);

    } else {
        // Duplicate or bad record. Free it.
        dbus_player_delete_item(player);
    }
}

void add_totem_player(gboolean find) {
    // Add Totem

    // ******************************
    // Totem player removed because it does now follow the MPRIS standard well.
    // ******************************
    return;


    gchar *service_name = "org.mpris.Totem";

    // Already in the list?
    if (dbus_player_lookup_service_name(service_name)) return;

    if (find) {
        // Find /usr/bin/totem
        gchar *path = find_command_path("totem");
        if (!path) {
            return;
        }
        g_free(path);
    }

    // New MediaPlayer record
    MediaPlayerRec *player = mpris_player_new(service_name);

    // Set real application name
    mpris_get_app_name(player);

    // Totem may not understand the "Identity" message!
    if (!player->app_name) {
        // Name of the Totem Media Player.
        player->app_name = g_strdup(_("Totem Media Player"));
    }

    // Function to connect/disconnect event signals
    player->func_set_signals = mpris_set_signals;

    // Function to get track-info (album, track/song name/title, genre, etc.)
    player->func_get_info = mpris_get_info;

    // Function to start/run the media player
    player->func_start_app = mpris_start_app;

    // Function to check if this player is up & running
    player->active = mpris_service_is_running_by_name(service_name);

    // Set icon
    player->icon_name = g_strdup("totem.png");

    if (!dbus_player_lookup_app_name(player->app_name)) {
        // Add to list
        GHashTable *player_list = dbus_player_get_list_ref();
        g_hash_table_insert(player_list, g_strdup(player->service_name), player);

    } else {
        // Duplicate or bad record. Free it.
        dbus_player_delete_item(player);
    }
}

void dbus_player_send_notification(MediaPlayerRec *player, gchar *msg) {
    MediaPlayerRec *p = NULL;
    if (player) {
        p = player;
    } else {
        // Create a "dummy" MediaPlayerRec
        p = mpris_player_new(NULL);
    }

    // Set active so the message goes through
    p->active = TRUE;

    // Send msg in the track field
    TrackInfo *tr = &p->track;

    str_copy(tr->track, msg, MPRIS_STRLEN-1);

    tr->status = PLAYER_STATUS_NOTIFY_MSG;

    // Sends the message via dbus-player.c (as all DBus modules should do!)
    dbus_player_process_data(p);

    // Delete the tmp player
    if (!player) {
        dbus_player_delete_item(p);
    }
}

void dbus_player_detect_software() {
    // Get the list of DBus services (software that i sregistered on the DBus).
    // Add all MPRIS (org.mpris.*) compliant programs to player list.

#define DBUS_MPRIS_NAMESPACE "org.mpris."

    // Connect to glib/DBus
    DBusGConnection *dbus_conn = dbus_player_connect_to_dbus();
    if (!dbus_conn) return;

    // Get player list
    GHashTable *player_list = dbus_player_get_list_ref();

    // Get service names from the DBus
    char **name_list = NULL;

    // Reset error structure
    GError *error = NULL;

    // These values come from:
    // http://dbus.freedesktop.org/doc/api/html/group__DBusShared.html
    //
    // DBUS_SERVICE_DBUS   "org.freedesktop.DBus"
    // DBUS_PATH_DBUS      "/org/freedesktop/DBus"
    // DBUS_INTERFACE_DBUS "org.freedesktop.DBus"
    DBusGProxy *proxy = dbus_g_proxy_new_for_name(dbus_conn,
                        DBUS_SERVICE_DBUS,
                        DBUS_PATH_DBUS,
                        DBUS_INTERFACE_DBUS);

    // Call ListNames method, wait for reply
    if (!dbus_g_proxy_call(proxy, "ListNames", &error, G_TYPE_INVALID, G_TYPE_STRV, &name_list, G_TYPE_INVALID)) {
        if (error->domain == DBUS_GERROR && error->code == DBUS_GERROR_REMOTE_EXCEPTION)
            LOG_ERROR("dbus_player_detect_software: Caught remote method exception %s: %s", dbus_g_error_get_name(error), error->message);
        else
            LOG_ERROR("dbus_player_detect_software: %s\n", error->message);

        g_error_free(error);
        return;
    }

    // Loop through gchar *name_list[]
    gchar **cptr;
    for (cptr = name_list; *cptr; cptr++) {
        // Drop names that start with ":"
        gchar *service_name = *cptr;

        if (service_name && service_name[0] == ':')  {
            continue;
        }

        // MediaPlayer record
        MediaPlayerRec *player = NULL;


        // Uncomment this to see all found service names
        // LOG_PLAYER("DBus service_name:%s\n", service_name);

        // Non-standard (non-MPRIS) media player
        if (!strncmp(service_name, "org.gnome.Rhythmbox", strlen("org.gnome.Rhythmbox"))) {

            add_rhythmbox_player(FALSE);

        }

        // RhythmBox is MPRIS2 compatible player
        else if (!strncmp(service_name, "org.mpris.MediaPlayer2.rhythmbox", strlen("org.mpris.MediaPlayer2.rhythmbox"))) {

            // TODO: Add support for org.mpris.MediaPlayer2.rhythmbox

        }

        // Non-standard (non-MPRIS) media player
        else if (!strncmp(service_name, "org.bansheeproject.Banshee", strlen("org.bansheeproject.Banshee"))) {

            add_banshee_player(FALSE);

        }

        // Somehow non-standard (non-MPRIS) media player
        // else if (!strncmp(service_name, "org.gnome.Totem", strlen("org.gnome.Totem"))) {
        else if (!strncmp(service_name, "org.mpris.Totem", strlen("org.mpris.Totem"))) {

            add_totem_player(FALSE);

        }

        // Skype
        else if (!g_strcmp0(service_name, "com.Skype.API")) {

            add_skype(FALSE);
        }

        // Add all standard following media players (org.mpris.*).
        // Eg. The Amarok 2.x player implements the MPRIS DBus API well.
        else if (!strncmp(DBUS_MPRIS_NAMESPACE, service_name, strlen(DBUS_MPRIS_NAMESPACE))) {

            // New MediaPlayer record
            player = mpris_player_new(service_name);

            // Set real application name (player->app_name)
            mpris_get_app_name(player);

            // TODO: Check this. Remove hard-coding if possible.
            // Set icon.
            if (g_strrstr(player->service_name, "amarok") || g_strrstr(player->service_name, "Amarok")) {
                player->icon_name = g_strdup("amarok.png");
            } else {
                player->icon_name = g_strdup("sound.png");
            }

            // Function to connect/disconnect event signals for this players
            player->func_set_signals = mpris_set_signals;

            // Function to check if this player is running
            player->func_check_is_running = mpris_service_is_running;

            // Function to get track-info (album, track/song name/title, genre, etc.)
            player->func_get_info = mpris_get_info;

            // Function to start/run the media player
            player->func_start_app = mpris_start_app;
        }

        // Has a player record?
        if (player && player->app_name) {

            // Is this player running/is active?
            player->active = mpris_service_is_running_by_name(player->service_name);

            // Add player to the g_player_list.

            // Some players send double set of information to the DBus.
            // Eg. "Amarok 2.3.2" identifies itself as service_name 'org.mpris.amarok' and 'org.mpris.MediaPlayer2.amarok'.

            // Lookup the record by its app_name (like: "Amarok 2.3.2")
            if (!dbus_player_lookup_app_name(player->app_name)) {
                // Add it to the list
                g_hash_table_insert(player_list, g_strdup(player->service_name), player);
            } else {
                // Duplicate or bad record. Free it.
                dbus_player_delete_item(player);
            }
        }
    }

    // Free the name_list
    g_strfreev(name_list);

    // Unref the proxy
    g_object_unref(proxy);

}


