/*
 * Copyright (c) Linux community.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
#include "dbus-player.h"
#include "dbus-mpris.h"
#include "dbus-rhythmbox.h"
#include "log.h"
#include "support.h"
#include "utility.h"

void rhythmbox_player_track_changed(MediaPlayerRec *player) {
    // Called when the track changes

    // Re-read all player data
    rhythmbox_get_info(player);

    TrackInfo *tr = &player->track;
    tr->status = PLAYER_STATUS_STOPPED;

    // Print debug info
    // dbus_player_debug_print(player);

    // Process data
    dbus_player_process_data(player);
}

void rhythmbox_player_status_changed(MediaPlayerRec *player) {
    // Called when player's status changes

    // Re-read all player data
    rhythmbox_get_info(player);

    // Print debug info
    // dbus_player_debug_print(player);

    // Process data
    dbus_player_process_data(player);
}

void rhythmbox_get_app_name(MediaPlayerRec *player) {
    // Name of the RhythmBox Media Player.
    player->app_name = g_strdup(_("RhythmBox Media Player"));
}

void rhythmbox_track_signal_cb(DBusGProxy *player_proxy, gpointer data, MediaPlayerRec *player) {
    // Track changed signal from RhythmBox

    gchar *track = (gchar*)data;
    if (track) {
        /*suppress unused variable msg*/
    }

    LOG_PLAYER("rhythmbox_track_signal_cb: %s (%s), TrackChanged: %s\n", player->app_name, player->service_name, track);

    rhythmbox_player_track_changed(player);
}

void rhythmbox_status_signal_cb(DBusGProxy *player_proxy, gpointer data, MediaPlayerRec *player) {
    // Status signal from RhythmBox

    gboolean status = (gint)(long)data;
    if (status) {
        /*suppress unused variable msg*/
    }
    // Pause=0
    // Playing=1

    LOG_PLAYER("rhythmbox_status_signal_cb: %s (%s) Status: %d\n", player->app_name, player->service_name, status);

    rhythmbox_player_status_changed(player);
}

gboolean rhythmbox_check_proxy(MediaPlayerRec *player) {
    // Check and set proxy
    if (!player) return FALSE;

    if (!player->proxy) {
        DBusGConnection *dbus_conn = dbus_player_connect_to_dbus();

        player->proxy = dbus_g_proxy_new_for_name(dbus_conn,
                        player->service_name,
                        "/org/gnome/Rhythmbox/Player",
                        "org.gnome.Rhythmbox.Player");
    }

    if (!player->proxy) {
        LOG_ERROR("rhythmbox_check_proxy: Cannot create DBus proxy for RhythmBox.\n");
    }

    return (player->proxy != NULL);
}

void rhythmbox_set_signals(gpointer player_rec, gboolean connect) {
    // Connect/disconnect signals

    MediaPlayerRec *player = (MediaPlayerRec*)player_rec;
    if (!player) return;

    if (!player->proxy) {

        // Check/set proxy
        if (!rhythmbox_check_proxy(player)) return;

        // Register arguments.
        dbus_g_proxy_add_signal(player->proxy, "playingUriChanged", G_TYPE_STRING, G_TYPE_INVALID);
        dbus_g_proxy_add_signal(player->proxy, "playingChanged", G_TYPE_BOOLEAN, G_TYPE_INVALID);
    }

    if (!rhythmbox_check_proxy(player)) return;

    if (connect) {
        // Connect signals
        dbus_g_proxy_connect_signal(player->proxy, "playingUriChanged", G_CALLBACK(rhythmbox_track_signal_cb), (gpointer)player, NULL);
        dbus_g_proxy_connect_signal(player->proxy, "playingChanged", G_CALLBACK(rhythmbox_status_signal_cb), (gpointer)player, NULL);
    } else  {
        // Disconnect signals
        dbus_g_proxy_disconnect_signal(player->proxy, "playingUriChanged", G_CALLBACK(rhythmbox_track_signal_cb), (gpointer)player);
        dbus_g_proxy_disconnect_signal(player->proxy, "playingChanged", G_CALLBACK(rhythmbox_status_signal_cb), (gpointer)player);
    }
}

gboolean get_hash_str(GHashTable *table, gchar *key, gchar *dest) {
    dest[0] = '\0';

    GValue* value = (GValue*)g_hash_table_lookup(table, key);

    if (value && G_VALUE_HOLDS_STRING(value)) {
        str_copy(dest, (gchar*)g_value_get_string(value), MPRIS_STRLEN-1);
        return TRUE;
    }

    return FALSE;
}

guint get_hash_uint(GHashTable *table, gchar *key) {
    guint ret = 0;
    GValue* value = (GValue*) g_hash_table_lookup(table, key);

    if (G_VALUE_HOLDS_UINT(value)) {
        ret = g_value_get_uint(value);
    }

    return ret;
}

void rhythmbox_get_info(gpointer player_rec) {
    MediaPlayerRec *player = (MediaPlayerRec*)player_rec;
    TrackInfo *tr = &player->track;

    tr->status = PLAYER_STATUS_STOPPED;

    // Check/set proxy
    if (!rhythmbox_check_proxy(player)) return;

    DBusGProxy *shell_proxy = NULL;

    // Is this player running/active?
    player->active = mpris_service_is_running_by_name(player->service_name);
    if (!player->active) {
        tr->status = PLAYER_STATUS_CLOSED;
        goto LBL_1;
    }

    DBusGConnection *dbus_conn = dbus_player_connect_to_dbus();

    GError *error = NULL;
    shell_proxy = dbus_g_proxy_new_for_name(dbus_conn,
                                            "org.gnome.Rhythmbox",
                                            "/org/gnome/Rhythmbox/Shell",
                                            "org.gnome.Rhythmbox.Shell");

    gboolean playing;
    if (!dbus_g_proxy_call_with_timeout(player->proxy, "getPlaying", DBUS_MPRIS_TIMEOUT, &error,
                                        G_TYPE_INVALID,
                                        G_TYPE_BOOLEAN, &playing,
                                        G_TYPE_INVALID)) {

        LOG_ERROR("rhythmbox_get_info: Failed to get playing state from RhythmBox. %s.", error ? error->message : "");
        tr->status = PLAYER_STATUS_STOPPED;

        if (error)
            g_error_free(error);

        goto LBL_1;
    }

    gchar *uri = NULL;
    if (!dbus_g_proxy_call_with_timeout(player->proxy, "getPlayingUri", DBUS_MPRIS_TIMEOUT, &error,
                                        G_TYPE_INVALID,
                                        G_TYPE_STRING, &uri,
                                        G_TYPE_INVALID)) {
        LOG_ERROR("rhythmbox_get_info: Failed to get song uri from RhythmBox. %s.", error ? error->message : "");

        if (error)
            g_error_free(error);

        goto LBL_1;
    }

    GHashTable *table = NULL;
    if (!dbus_g_proxy_call_with_timeout(shell_proxy, "getSongProperties", DBUS_MPRIS_TIMEOUT, &error, G_TYPE_STRING, uri, G_TYPE_INVALID,
                                        dbus_g_type_get_map("GHashTable", G_TYPE_STRING, G_TYPE_VALUE),	&table, G_TYPE_INVALID)) {

        if (!playing) {
            tr->status = PLAYER_STATUS_STOPPED;
        }

        if (error)
            g_error_free(error);

        goto LBL_1;
    }

    g_free(uri);

    if (playing)
        tr->status = PLAYER_STATUS_PLAYING;
    else
        tr->status = PLAYER_STATUS_PAUSED;

    // Check if stream title is nonempty, if so, take title
    if (!get_hash_str(table, "rb:stream-song-title", tr->track)) {
        get_hash_str(table, "title", tr->track);
    }

    get_hash_str(table, "artist", tr->artist);
    get_hash_str(table, "album", tr->album);

    tr->total_secs = get_hash_uint(table, "duration");

    g_hash_table_destroy(table);

    if (!dbus_g_proxy_call_with_timeout(player->proxy, "getElapsed", DBUS_MPRIS_TIMEOUT, &error,
                                        G_TYPE_INVALID,
                                        G_TYPE_UINT, &tr->current_secs,
                                        G_TYPE_INVALID)) {

        LOG_ERROR("rhythmbox_get_info: Failed to get elapsed time from RhythmBox. %s.", error ? error->message : "");

        if (error)
            g_error_free(error);
    }

LBL_1:
    // Unref proxy
    if (shell_proxy)
        g_object_unref(shell_proxy);

}

void rhythmbox_start_app(gpointer player_rec) {
    // Function to start RhythmBox
    MediaPlayerRec *player = (MediaPlayerRec*)player_rec;

    LOG_PLAYER("rhythmbox_start_app: %s\n", player->app_name);

    if (!rhythmbox_check_proxy(player)) return;

    // Start player
    dbus_g_proxy_call_no_reply(player->proxy, "Play", G_TYPE_INVALID);
}

