/*
 * Copyright (c) Linux community.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
*/
#include <string.h>
#include "dconf.h"
#include "support.h"
#include "utility.h"
#include "log.h"

// This module writes and reads values from/to GNOME's configuration registry (dconf registry).
// You can check these values in dconf-editor.
// Start dconf-editor and browse to /apps/audio-recorder/.

// Notice:
// All valid configuration keys must be defined in the schema file "org.gnome.audio-recorder.gschema.xml".
// It's in the data/ folder.
// You must also compile and install this file to /usr/share/glib-2.0/schemas/ directory.
//
// 0) Test the schema file first. I assume here that the schema is in the current (.) directory.
//    $ glib-compile-schemas --dry-run .
//
// 1) Copy the schema file to /usr/share/glib-2.0/schemas/.
//    $ sudo cp org.gnome.audio-recorder.gschema.xml /usr/share/glib-2.0/schemas/
//
// 2) Then re-compile all schemas.
//    $ sudo glib-compile-schemas /usr/share/glib-2.0/schemas/
//
// The Makefile.am in the data/ folder should do this installation automatically.
//
// You can also study various schema values with gsettings tool.
// $ gsettings
// -----------------------------------------------------------------
//
// The main schema name is "org.gnome.audio-recorder", and its path is /apps/audio-recorder/.
// Notice that there are also child schemas (sub paths) for
//  track/ (/apps/audio-recorder/track/).
//  skype/ (/apps/audio-recorder/skype/).
//  players/ (/apps/audio-recorder/players/).

// The main configuration schema for this application
#define APPLICATION_SETTINGS_SCHEMA "org.gnome.audio-recorder"

gboolean conf_is_valid_key(GSettings *settings, gchar *key) {
    // Check if the key is valid key within settings
    if (!G_IS_SETTINGS(settings)) {
        return FALSE;
    }

    // Get list of keys
    gchar **keys = NULL;
    keys = g_settings_list_keys(settings);

    gint i = 0;
    gboolean found = FALSE;
    while (keys && keys[i]) {
        if (!g_strcmp0(key, keys[i])) {
            found = TRUE;
            break;
        }
        i++;
    }

    // Free keys
    g_strfreev(keys);

    return found;
}

#if 0
void conf_list_keys(GSettings *settings) {
    // List keys for the given settings (and schema)
    gchar *str = NULL;
    g_object_get(settings, "schema", &str, NULL);
    if (!str) {
        g_print("Bad schema:%s\n", str);
    }

    gchar **keys = NULL;
    gint i = 0;
    keys = g_settings_list_keys(settings);
    while (keys && keys[i]) {
        g_print("List key for schema:%s key:%s\n",  str, keys[i]);
        i++;
    }

    g_strfreev(keys);
    g_free(str);
}

void conf_list_children(GSettings *settings) {
    // List children of the given settings (and schema)
    gchar *str = NULL;
    g_object_get(settings, "schema", &str, NULL);
    if (!str) {
        g_print("Bad schema:%s\n", str);
    }

    gchar **children = NULL;
    gint i = 0;
    children = g_settings_list_children(settings);
    while (children && children[i]) {
        g_print("List child for schema:%s child:%s\n",  str, children[i]);
        i++;
    }

    g_strfreev(children);
    g_free(str);
}
#endif

void conf_get_child_path(gchar *key, gchar **child_path, gchar **child_key) {
    // Split the key to child_path and child_key.
    // Eg. "track/track-name" has child_path "track/" and child_key "track-name".
    *child_path = NULL;
    *child_key = NULL;

    static gchar buf[MAX_PATH_LEN];

    // Find "/" in key
    gchar *p = g_strrstr(key, "/");
    if (p) {
        memset(buf, '\0', MAX_PATH_LEN);
        g_utf8_strncpy(buf, key, p - key);
        *child_path = g_strdup(buf);

        *child_key = g_strdup(p+1);
    }
}

GSettings *conf_get_base_settings() {
    // Return GSettings (base) object. This points to /apps/audio-recorder/.
    GSettings *settings = g_settings_new(APPLICATION_SETTINGS_SCHEMA);
    if (!G_IS_SETTINGS(settings)) {
        LOG_ERROR("Cannot find settings for %s\n", APPLICATION_SETTINGS_SCHEMA);
        return NULL;
    }
    return settings;
}

GSettings *conf_get_settings_for_key(gchar *key, gchar **child_path, gchar **child_key) {
    // Return GSettings object for the given key
    *child_path = NULL;
    *child_key = NULL;

    // The key may contain a child path (it has a "/" in it).
    // For example the key "track/track-name" contains child_path ("track/") and child_key "track-name".
    // Take possible child_path and child_key.
    conf_get_child_path(key, child_path, child_key);

    // The main GSettings object
    GSettings *settings = conf_get_base_settings(); // Points to /apps/audio-recorder/.
    if (!G_IS_SETTINGS(settings)) {
        return NULL;
    }

    if (*child_path) {
        // Get GSettings object for child_path
        GSettings *child_settings = g_settings_get_child(settings, *child_path);
        if (G_IS_SETTINGS(child_settings)) {
            g_object_unref(settings);
            settings = child_settings; // Points to /apps/audio-recorder/some child_path/.
        }
    }

    return settings;
}

void conf_get_boolean_value(gchar *key, gboolean *value) {
    gchar *child_path = NULL;
    gchar *child_key = NULL;

    // Get GSettings object. Handle also child_path and child_key (like "track/track-name")
    GSettings *settings = conf_get_settings_for_key(key, &child_path, &child_key);

    gchar *k = NULL;
    if (child_key)
        k = child_key; // settings object points to child_path (/apps/audio-recorder/some child_path/).
    else
        k = key; // settings object points to main path (/apps/audio-recorder/).

    // Check if the key is valid. Avoid crash.
    if (!conf_is_valid_key(settings, k)) {
        LOG_ERROR("Cannot find configuration key \"%s\".\n", key);
        goto LBL_1;
    }

    // Read value
    *value = g_settings_get_boolean(settings, k);

LBL_1:
    // Free values
    g_free(child_path);
    g_free(child_key);
    g_object_unref(settings);
}

void conf_get_int_value(gchar *key, gint *value) {
    gchar *child_path = NULL;
    gchar *child_key = NULL;

    // Get GSettings object. Handle also child_path and child_key (like "track/track-name")
    GSettings *settings = conf_get_settings_for_key(key, &child_path, &child_key);

    gchar *k = NULL;
    if (child_key)
        k = child_key; // settings object points to child_path (/apps/audio-recorder/some child_path/).
    else
        k = key; // settings object points to main path (/apps/audio-recorder/).

    // Check if the key is valid. Avoid crash.
    if (!conf_is_valid_key(settings, k)) {
        LOG_ERROR("Cannot find configuration key \"%s\".\n", key);
        goto LBL_1;
    }

    // Read value
    *value = g_settings_get_int(settings, k);

LBL_1:
    // Free values
    g_free(child_path);
    g_free(child_key);
    g_object_unref(settings);
}

void conf_get_string_value(gchar *key, gchar **value) {
    gchar *child_path = NULL;
    gchar *child_key = NULL;

    // Get GSettings object. Handle also child_path and child_key (like "track/track-name")
    GSettings *settings = conf_get_settings_for_key(key, &child_path, &child_key);

    gchar *k = NULL;
    if (child_key)
        k = child_key; // settings object points to child_path (/apps/audio-recorder/some child_path/).
    else
        k = key; // settings object points to main path (/apps/audio-recorder/).

    // Check if the key is valid. Avoid crash.
    if (!conf_is_valid_key(settings, k)) {
        LOG_ERROR("Cannot find configuration key \"%s\".\n", key);
        goto LBL_1;
    }

    // Read value
    *value = g_settings_get_string(settings, k);

LBL_1:
    // Free values
    g_free(child_path);
    g_free(child_key);
    g_object_unref(settings);

    // The caller should g_free() the value
}

void conf_get_string_list(gchar *key, GList **list) {
    gchar *child_path = NULL;
    gchar *child_key = NULL;

    // Get GSettings object. Handle also child_path and child_key (like "track/track-name")
    GSettings *settings = conf_get_settings_for_key(key, &child_path, &child_key);

    gchar *k = NULL;
    if (child_key)
        k = child_key; // settings object points to child_path (/apps/audio-recorder/some child_path/).
    else
        k = key; // settings object points to main path (/apps/audio-recorder/).

    // Check if the key is valid. Avoid crash.
    if (!conf_is_valid_key(settings, k)) {
        LOG_ERROR("Cannot find configuration key \"%s\".\n", key);
        goto LBL_1;
    }

    // Read string list
    gchar **argv = g_settings_get_strv(settings, k);

    // From gchar *argv[] to GList
    *list = NULL;
    guint i = 0;
    while (argv && argv[i]) {
        *list = g_list_append(*list, g_strdup(argv[i]));
        i++;
    }

    // Free argv[]
    g_strfreev(argv);


LBL_1:
    // Free values
    g_free(child_path);
    g_free(child_key);
    g_object_unref(settings);

    // The caller should free the list
}

void conf_save_boolean_value(gchar *key, gboolean value) {
    gchar *child_path = NULL;
    gchar *child_key = NULL;

    // Get GSettings object. Handle also child_path and child_key (like "track/track-name")
    GSettings *settings = conf_get_settings_for_key(key, &child_path, &child_key);

    gchar *k = NULL;
    if (child_key)
        k = child_key; // settings object points to child_path (/apps/audio-recorder/some child_path/).
    else
        k = key; // settings object points to main path (/apps/audio-recorder/).


    // Check if the key is valid. Avoid crash.
    if (!conf_is_valid_key(settings, k)) {
        LOG_ERROR("Cannot find configuration key \"%s\".\n", key);
        goto LBL_1;
    }

    // Save value
    if (!g_settings_set_boolean(settings, k, value)) {
        LOG_ERROR("Cannot save configuration key \"%s\" (%s).\n", key, (value ? "true" : "false"));
    }

LBL_1:
    // Free values
    g_free(child_path);
    g_free(child_key);
    g_object_unref(settings);
}

void conf_save_int_value(gchar *key, gint value) {
    gchar *child_path = NULL;
    gchar *child_key = NULL;

    // Get GSettings object. Handle also child_path and child_key (like "track/track-name")
    GSettings *settings = conf_get_settings_for_key(key, &child_path, &child_key);

    gchar *k = NULL;
    if (child_key)
        k = child_key; // settings object points to child_path (/apps/audio-recorder/some child_path/).
    else
        k = key; // settings object points to main path (/apps/audio-recorder/).

    // Check if the key is valid. Avoid crash.
    if (!conf_is_valid_key(settings, k)) {
        LOG_ERROR("Cannot find configuration key \"%s\".\n", key);
        goto LBL_1;
    }

    // Save value
    if (!g_settings_set_int(settings, k, value)) {
        LOG_ERROR("Cannot save configuration key \"%s\" (%d).\n", key, value);
    }

LBL_1:
    // Free values
    g_free(child_path);
    g_free(child_key);
    g_object_unref(settings);
}

void conf_save_string_value(gchar *key, gchar *value) {
    gchar *child_path = NULL;
    gchar *child_key = NULL;

    // Get GSettings object. Handle also child_path and child_key (like "track/track-name")
    GSettings *settings = conf_get_settings_for_key(key, &child_path, &child_key);

    gchar *k = NULL;
    if (child_key)
        k = child_key; // settings object points to child_path (/apps/audio-recorder/some child_path/).
    else
        k = key; // settings object points to main path (/apps/audio-recorder/).

    // Check if the key is valid. Avoid crash.
    if (!conf_is_valid_key(settings, k)) {
        LOG_ERROR("Cannot find configuration key \"%s\".\n", key);
        goto LBL_1;
    }

    // Save value
    if (!g_settings_set_string(settings, k, value)) {
        LOG_ERROR("Cannot save configuration key \"%s\" (%s).\n", key, value);
    }

LBL_1:
    // Free values
    g_free(child_path);
    g_free(child_key);
    g_object_unref(settings);
}

void conf_save_string_list(gchar *key, GList *list) {
    gchar *child_path = NULL;
    gchar *child_key = NULL;

    // Get GSettings object. Handle also child_path and child_key (like "track/track-name")
    GSettings *settings = conf_get_settings_for_key(key, &child_path, &child_key);

    gchar *k = NULL;
    if (child_key)
        k = child_key; // settings object points to child_path (/apps/audio-recorder/some child_path/).
    else
        k = key; // settings object points to main path (/apps/audio-recorder/).

    // Check if the key is valid. Avoid crash.
    if (!conf_is_valid_key(settings, k)) {
        LOG_ERROR("Cannot find configuration key \"%s\".\n", key);
        goto LBL_1;
    }

    // From GList to gchar *argv[].
    guint len = g_list_length(list);
    // Allocate argv[]
    gchar **argv = g_new(gchar*, len + 1 /*+1 for NULL*/);

    guint i = 0;
    GList *n = g_list_first(list);
    while (n) {
        argv[i++] = g_strdup(n->data);
        n = g_list_next(n);
    }
    argv[len] = NULL;

    // Save string list
    if (!g_settings_set_strv(settings, k, (const gchar* const*)argv)) {
        LOG_ERROR("Cannot save configuration key \"%s\" (value is a string list).\n", key);
    }

    // Free argv[]
    g_strfreev((gchar **)argv);

LBL_1:
    // Free values
    g_free(child_path);
    g_free(child_key);
    g_object_unref(settings);
}


#if 0

GCONF:
This is the old GConf2 based solution:

gchar *conf_get_app_name() {
    // Caller should g_free() this value
    return g_strdup(APPLICATION_KEY);
}

gchar *conf_get_key_path(gchar *key) {
    gchar *app = conf_get_app_name();
    gchar *key_path = g_strdup_printf("/apps/%s/%s", app, key);
    g_free(app);
    return key_path;
}

void conf_save_boolean_value(gchar *key, gboolean value) {
    GConfClient *gc_client = gconf_client_get_default();
    g_return_if_fail(GCONF_IS_CLIENT(gc_client));

    gchar *key_path = conf_get_key_path(key);

    if(!gconf_client_set_bool(gc_client, key_path, value, NULL)) {
        LOG_WARNING("gconf.c: Failed to set key %s (%s)\n", key_path, (value ? "True" : "False") );
    }

    g_free(key_path);
    g_object_unref(gc_client);
}

void conf_save_int_value(gchar *key, gint value) {
    GConfClient *gc_client = gconf_client_get_default();
    g_return_if_fail(GCONF_IS_CLIENT(gc_client));

    gchar *key_path = conf_get_key_path(key);

    if(!gconf_client_set_int(gc_client, key_path, value, NULL)) {
        LOG_WARNING("gconf.c: Failed to set key %s (%d)\n", key_path, value);
    }

    g_free(key_path);
    g_object_unref(gc_client);
}

void conf_save_string_value(gchar *key, gchar *value) {
    GConfClient *gc_client = gconf_client_get_default();
    g_return_if_fail(GCONF_IS_CLIENT(gc_client));

    gchar *key_path = conf_get_key_path(key);

    if(!gconf_client_set_string(gc_client, key_path, value, NULL)) {
        LOG_WARNING("gconf.c: Failed to set key %s (%s)\n", key_path, value);
    }

    g_free(key_path);
    g_object_unref(gc_client);
}

#if 0
static void conf_free_value_list(GSList **list) {
    GSList *sitem = *list;
    while (sitem) {
        GConfValue *value = sitem->data;
        gconf_value_free(value);
        sitem = sitem->next;
    }
    g_slist_free(*list);
    *list = NULL;
}
#endif

void conf_save_string_list(gchar *key, GList *list) {
    // Notice: gconf_client_set_list() takes GSList.
    // Copy GList to GSList.
    GSList *slist = NULL;
    GList *n = g_list_first(list);
    while (n) {
        gchar *s = (gchar*)n->data;
        slist = g_slist_append(slist, s/*steal the string*/);
        n = g_list_next(n);
    }

    GConfClient *gc_client = gconf_client_get_default();
    g_return_if_fail(GCONF_IS_CLIENT(gc_client));

    gchar *key_path = conf_get_key_path(key);

    if(!gconf_client_set_list(gc_client, key_path, GCONF_VALUE_STRING, slist/*GSList*/, NULL)) {
        LOG_WARNING("gconf.c: Failed to save list to key %s.\n", key_path);
    }

    // Free the slist. Do not free data.
    g_slist_free(slist);

    g_free(key_path);
    g_object_unref(gc_client);
}

/*
void conf_save_string_list(gchar *key, GSList *list) {
    GConfClient *gc_client = gconf_client_get_default();
    g_return_if_fail(GCONF_IS_CLIENT(gc_client));

    gchar *key_path = conf_get_key_path(key);

    if(!gconf_client_set_list(gc_client, key_path, GCONF_VALUE_STRING, list, NULL)) {
        LOG_WARNING("gconf.c: Failed to save list to key %s.\n", key_path);
    }

    g_free(key_path);
    g_object_unref(gc_client);
}
*/

gboolean conf_get_boolean_value(gchar *key, gboolean *value) {
    *value = FALSE;
    gboolean ret = FALSE;

    GConfClient *gc_client = gconf_client_get_default();
    g_return_val_if_fail(GCONF_IS_CLIENT(gc_client), FALSE);

    gchar *key_path = conf_get_key_path(key);

    GConfValue *gc_val = NULL;
    gc_val = gconf_client_get_without_default(gc_client, key_path, NULL);

    /* If value pointer remains NULL, the key was not found */
    if(gc_val == NULL) {
        LOG_WARNING("gconf.c: Cannot find key %s.\n", key_path);
        goto l_end;
    }

    /* Check if the value type is boolean */
    if(gc_val->type == GCONF_VALUE_BOOL) {
        *value = gconf_value_get_bool(gc_val);
    }

    ret = TRUE;

l_end:
    g_free(key_path);

    if (gc_val)
        gconf_value_free(gc_val);

    g_object_unref(gc_client);

    return ret;
}

gboolean conf_get_int_value(gchar *key, gint *value) {
    *value = 0L;
    gboolean ret = FALSE;

    GConfClient *gc_client = gconf_client_get_default();
    g_return_val_if_fail(GCONF_IS_CLIENT(gc_client), FALSE);

    gchar *key_path = conf_get_key_path(key);

    GConfValue *gc_val = NULL;
    gc_val = gconf_client_get_without_default(gc_client, key_path, NULL);

    if(gc_val == NULL) {
        LOG_WARNING("gconf.c: Cannot find key %s.\n", key_path);
        goto l_end;
    }

    /* Check if the value type is integer */
    if(gc_val->type == GCONF_VALUE_INT) {
        *value = gconf_value_get_int(gc_val);
    }
    ret = TRUE;

l_end:
    g_free(key_path);
    if (gc_val)
        gconf_value_free(gc_val);
    g_object_unref(gc_client);
    return ret;
}

gboolean conf_get_string_value(gchar *key, gchar **value) {
    *value = NULL;
    gboolean ret = FALSE;

    GConfClient *gc_client = gconf_client_get_default();
    g_return_val_if_fail(GCONF_IS_CLIENT(gc_client), FALSE);

    gchar *key_path = conf_get_key_path(key);

    GConfValue *gc_val = NULL;
    gc_val = gconf_client_get_without_default(gc_client, key_path, NULL);

    if(gc_val == NULL) {
        LOG_WARNING("gconf.c: Cannot find key %s.\n", key_path);
        goto l_end;
    }

    /* Check if the value type is string */
    if(gc_val->type == GCONF_VALUE_STRING) {
        *value = g_strdup(gconf_value_get_string(gc_val));
    }
    ret = TRUE;

l_end:
    g_free(key_path);
    if (gc_val)
        gconf_value_free(gc_val);
    g_object_unref(gc_client);
    return ret;
}

gboolean conf_get_string_list(gchar *key, GList **list) {
    *list = NULL;
    gboolean ret = FALSE;

    GConfClient *gc_client = gconf_client_get_default();
    g_return_val_if_fail(GCONF_IS_CLIENT(gc_client), FALSE);

    gchar *key_path = conf_get_key_path(key);

    GSList *slist = gconf_client_get_list(gc_client, key_path, GCONF_VALUE_STRING, NULL);

    // Notice: gconf_client_get_list() returns GSList.
    // Copy GSList to GList.
    GSList *n = slist;
    while (n) {
        gchar *s = (gchar*)n->data;
        *list = g_list_append(*list, s/*steal the string*/);
        n = g_slist_next(n);
    }

    if (!(*list)) {
        LOG_WARNING("gconf.c: Failed to get list for key %s.\n", key_path);
        goto l_end;
    }
    ret = TRUE;

l_end:
    g_object_unref(gc_client);
    g_free(key_path);

    // Free the GSlist, Do not free data, it's now owned by the *list.
    g_slist_free(slist);

    return ret;
}

#endif



