/*
 * Copyright (c) Linux community.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
*/
#include "gst-listener.h"
#include "audio-sources.h"
#include "dconf.h"
#include "log.h"
#include "utility.h"

#include "gst-pipeline.h"

#include <string.h>
#include <glib.h>
#include <gdk/gdk.h>
#include <math.h>
#include <gst/gst.h>

// Average level value (rms dB)
G_LOCK_DEFINE_STATIC(g_avg_rms_dB);
static gdouble g_avg_rms_dB = -INFINITY;
static GstClockTime g_last_time = GST_CLOCK_TIME_NONE;

// GStreamer pipeline
static GstElement *g_listener_pipeline;

// Debug flag
static gboolean g_debug_flag = FALSE;

// Current audio source and device list
static gchar *g_curr_audio_source = NULL;
static GList *g_curr_dev_list = NULL;

void listener_module_init();
void listener_module_exit();

static void listener_set_data(GstClockTime end_time, double rms_dB, double peak_dB, double decay_dB);

static GstElement *listener_create_pipeline(gchar *audio_source, GList *dev_list);
static void listener_save_device(gchar *audio_source, GList *dev_list);
static gboolean listener_is_listening();
static void listener_shutdown_pipeline();

void listener_module_init() {
    LOG_DEBUG("Init gst-listener.c.\n");

    // Initialize
    g_listener_pipeline = NULL;

    g_curr_audio_source = NULL;
    g_curr_dev_list = NULL;

    // Reset data record
    listener_clear_data();

    g_debug_flag = FALSE;
}

void listener_module_exit() {
    LOG_DEBUG("Clean up gst-listener.c.\n");

    listener_stop_listening();
}

void listener_set_debug_flag(gboolean on_off) {
    // Set debug flag. Please see application options:
    // $ audio-recorder --help
    // Listener process will now print level values in a terminal window.
    g_debug_flag = on_off;
}

void listener_save_device(gchar *audio_source, GList *dev_list) {
    // Save audio source
    if (g_curr_audio_source)
        g_free(g_curr_audio_source);

    g_curr_audio_source = audio_source;

    // Save device list
    str_list_free(g_curr_dev_list);
    g_curr_dev_list = dev_list;
}

void listener_start_listening() {
    // Get audio source and device list
    gchar *audio_source = NULL;
    GList *dev_list = audio_sources_get_device_NEW(&audio_source);

    // Values have changed since last check?
    gboolean changed = g_strcmp0(g_curr_audio_source, audio_source);
    changed = changed || (!str_lists_equal(g_curr_dev_list, dev_list));

    // Changed?
    if (changed) {
        // Yes, stop the listener.
        listener_stop_listening();
    }

    // Already listening (or values were unchanged)?
    if (listener_is_listening()) {
        // Free values
        g_free(audio_source);
        str_list_free(dev_list);
        return;
    }

    // Start listener with new parameters.
    // Create GStreamer pipeline and gather signal levels.
    g_listener_pipeline = listener_create_pipeline(audio_source, dev_list);

    // Save values
    listener_save_device(audio_source, dev_list);
}

void listener_stop_listening() {
    listener_shutdown_pipeline();

    // Clear variables
    listener_save_device(NULL, NULL);
}

static gboolean listener_is_listening() {
    // The pipeline has been created?
    if (!GST_IS_OBJECT(g_listener_pipeline)) return FALSE;

    // The state is?
    // Ref: http://www.gstreamer.net/data/doc/gstreamer/head/gstreamer/html/GstElement.html#gst-element-get-state
    GstState status, pending;
    gst_element_get_state(g_listener_pipeline, &status, &pending, 0);

    return (status != GST_STATE_NULL);
}

void listener_clear_data() {
    // Lock
    G_LOCK(g_avg_rms_dB);

    // Reset values
    g_last_time = GST_CLOCK_TIME_NONE;
    g_avg_rms_dB = -INFINITY;

    // Unlock
    G_UNLOCK(g_avg_rms_dB);
}

void listener_get_average_rms(gdouble *avg_rms_dB, gdouble *avg_rms) {
    // Lock
    G_LOCK(g_avg_rms_dB);

    // Take average rms (in decibels)
    gdouble avg_dB = g_avg_rms_dB;

    // From decibel to normalized 0 - 1.0
    gdouble avg = pow(10, avg_dB / 20);

    // To 0 - 100%
    avg = avg * 100.0;

    // Value is ok?
    if (isnan(avg_dB)) {
        avg_dB = -INFINITY;
        avg = 0;
    }

    // Return values
    *avg_rms_dB = avg_dB;
    *avg_rms = avg;

    // Unlock
    G_UNLOCK(g_avg_rms_dB);
}

static void listener_set_data(GstClockTime end_time, double rms_dB, double peak_dB, double decay_dB) {
    // Save level values.
    // Please study:
    // http://www.gstreamer.net/data/doc/gstreamer/head/gst-plugins-good-plugins/html/gst-plugins-good-plugins-level.html
    // http://en.wikipedia.org/wiki/Decibel

    // Set lock
    G_LOCK(g_avg_rms_dB);

    if (rms_dB == -INFINITY) {
        ;
    }

    // rms_dB:
    // RMS, http://en.wikipedia.org/wiki/Root_mean_square
    // From dB to a normalized 0 - 1.0 value.
    gdouble rms = pow(10, rms_dB / 20);

    // peak_dB:
    // Signal peak from dB to 0 - 1.0. Not used.
    double peak = exp(peak_dB / 20);
    if (peak > 1.0) peak = 1.0;

    // Ref: http://gstreamer.freedesktop.org/data/doc/gstreamer/head/gstreamer/html/GstClock.html
    if (!GST_CLOCK_TIME_IS_VALID(g_last_time)) {
        // First time
        g_avg_rms_dB = rms_dB;

    } else {
        // Calculate instant average (last avg. value + current)/2
        g_avg_rms_dB = (g_avg_rms_dB + rms_dB) / 2.0;
    }

    // Update timestamp
    g_last_time = end_time;

    // Recover from NAN
    if (isnan(g_avg_rms_dB) || g_avg_rms_dB == -INFINITY) {
        g_avg_rms_dB = rms_dB;
    }

    // Print debug info in a terminal window?
    // Audio-recorder was started with --debug-signal=1 or -d 1 options.
    // Please see:
    // $ audio-recorder --help
    if (g_debug_flag) {

        // Calculate normalized average (0 - 1.0)
        gdouble avg_rms = pow(10, g_avg_rms_dB / 20);

        LOG_MSG("Time: %" GST_TIME_FORMAT ": rms=%3.1fdB (=%3.1f%%) peak=%3.1fdB (=%3.1f%%) decay=%3.1fdB. Avg.rms=%3.1fdB (=%3.1f%%)\n",
                GST_TIME_ARGS(end_time), rms_dB, rms*100.0, peak_dB, peak*100.0, decay_dB, g_avg_rms_dB, avg_rms*100.0/*=% */);
    }

    // Unlock
    G_UNLOCK(g_avg_rms_dB);
}

static gboolean listener_message_handler(GstBus * bus, GstMessage * message, gpointer data) {
    if (message->type == GST_MESSAGE_ELEMENT) {

        const GstStructure *s = gst_message_get_structure(message);
        const gchar *name = gst_structure_get_name(s);

        if (strcmp (name, "level") == 0) {
            GstClockTime endtime;

            if (!gst_structure_get_clock_time(s, "endtime", &endtime)) {
                g_warning("Could not parse endtime");
            }

            // We can get the number of channels as the length of any of the value lists
            const GValue *list;
            list = gst_structure_get_value(s, "rms");

            gint channels = 0;
            channels = gst_value_list_get_size(list);

            gint i = 0;
            for (i = 0; i < channels; ++i) {
                const GValue *value;

                list = gst_structure_get_value(s, "rms");

                value = gst_value_list_get_value(list, i);
                gdouble rms_dB = g_value_get_double(value);

                list = gst_structure_get_value(s, "peak");
                value = gst_value_list_get_value (list, i);
                gdouble peak_dB = g_value_get_double(value);

                list = gst_structure_get_value(s, "decay");
                value = gst_value_list_get_value(list, i);
                gdouble decay_dB = g_value_get_double(value);

                listener_set_data(endtime, rms_dB, peak_dB, decay_dB);

            }
        }
    }

    // We handled the message we want, and ignored the ones we didn't want. so the core can unref the message for us
    return TRUE;
}

static void listener_shutdown_pipeline() {
    // Shutdown the listener pipeline

    if (!GST_IS_PIPELINE(g_listener_pipeline)) {
        goto LBL_1;
    }

    LOG_LISTENER("Shutdown listener pipeline.\n");

    // Switch to GST_STATE_NULL
    gst_element_set_state(g_listener_pipeline, GST_STATE_NULL);

    // Then destroy it
    gst_object_unref(GST_OBJECT(g_listener_pipeline));
    g_listener_pipeline = NULL;

LBL_1:
    listener_clear_data();
}

static GstElement *listener_create_pipeline(gchar *audio_source, GList *dev_list) {

#if defined(DEBUG_LISTENER) || defined(DEBUG_ALL)
    LOG_LISTENER("Start listener for \"%s\"\n", audio_source);
    str_list_print("Listen to devices", dev_list);
#endif

    gchar *err_msg = NULL;

    // Create pipeline
    gchar *encoding = "audio/x-raw-int,channels=1";
    GstElement *pipeline = pipeline_create(audio_source, dev_list, encoding, "fakesink", &err_msg);

    // Errors?
    if (!GST_IS_PIPELINE(pipeline) || err_msg) {
        goto LBL_1;
    }

    // Add message handler
    GstBus *bus = gst_pipeline_get_bus(GST_PIPELINE(pipeline));
    gst_bus_add_signal_watch(bus);

    // Monitor sound level/amplitude
    g_signal_connect(bus, "message::element", G_CALLBACK(listener_message_handler), NULL);

    // Sync
    // g_object_set(G_OBJECT(fake_sink), "sync", TRUE, NULL);

    gst_object_unref(bus);

    // Set the pipeline to "playing" state
    if (gst_element_set_state(pipeline, GST_STATE_PLAYING) == GST_STATE_CHANGE_FAILURE) {
        err_msg = g_strdup(_("Cannot start reading from the stream/pipeline.\n"));
        goto LBL_1;

    } else {
        LOG_DEBUG("Listener pipeline is OK.\n");
    }

    // Ok
    return pipeline;

LBL_1:
    // Error occured
    LOG_ERROR(err_msg);

    // Shutdown
    if (GST_IS_OBJECT(pipeline)) {
        LOG_LISTENER("Distroying listener pipeline.\n");

        // Switch to GST_STATE_NULL
        gst_element_set_state(pipeline, GST_STATE_NULL);

        // Then destroy it
        gst_object_unref(GST_OBJECT(pipeline));
    }
    pipeline = NULL;

    g_free(err_msg);

    // Show an error message
    LOG_ERROR("Cannot create listener for %s.\n",  audio_source);
    GList *n = g_list_first(dev_list);
    while (n) {
        gchar *device_id = (gchar*)n->data;
        LOG_ERROR("Listener device: %s\n", device_id);
        n = g_list_next(n);
    }

    return NULL;
}

